from __future__ import print_function

from os import environ
from decimal import Decimal

from .errors import Error
from .version import version


def check_parens(s):
    return s.count("(") == s.count(")")


def print_traceback(e):
    from .object import Object

    if e.args and isinstance(e.args[0], Object):
        error = e.args[0]
        type = str(error["type"]) if error["type"] is not None else error.type
        message = str(error["message"]) if error["message"] is not None else ""
    else:
        type = e.__class__.__name__
        message = str(e)

    stack = "\n".join(["  {0:s}".format(repr(m)) for m in e.stack])
    underline = "-" * (len(type) + 1)
    print("\n  {0:s}: {1:s}\n  {2:s}\n{3:s}\n".format(type, message, underline, stack))


class Completer(object):

    def __init__(self, state):
        super(Completer, self).__init__()

        self.state = state

        # Build up Global Objects
        self.objects = self.state.root.attrs.keys()
        self.objects.extend(self.state.root["Types"].attrs.keys())
        self.objects.extend(self.state.root["Core"].attrs.keys())
        if "builtins" in self.state.root:
            self.objects.extend(self.state.root["builtins"].attrs.keys())

    def complete(self, text, state):
        if state == 0:
            if " " in text:
                self.matches = self.attr_matches(text)
            else:
                self.matches = self.global_matches(text)

        try:
            return self.matches[state]
        except IndexError:
            return None

    def display_matches(self, substitution, matches, longest_match_length):
        import readline

        line_buffer = readline.get_line_buffer()
        columns = environ.get("COLUMNS", 80)

        print()

        tpl = "{:<" + str(int(max(map(len, matches)) * 1.2)) + "}"

        buffer = ""
        for match in matches:
            match = tpl.format(match[len(substitution):])
            if len(buffer + match) > columns:
                print(buffer)
                buffer = ""
            buffer += match

        if buffer:
            print(buffer)

        print(self.state.prompt, end="")
        print(line_buffer, end="")

    def global_matches(self, text):
        matches = []
        n = len(text)
        for word in self.objects:
            if word[:n] == text:
                matches.append(word)
        return matches

    def attr_matches(self, text):
        tokens = text.split(" ")
        expr, attr = " ".join(tokens[:-1]), tokens[-1]

        try:
            object = self.state.eval(expr, showerror=False, reraise=True)
        except Exception:
            return []

        words = object.attrs.keys()

        matches = []
        n = len(attr)
        for word in words:
            if word[:n] == attr:
                matches.append("{0:s} {1:s}".format(expr, word))
        return matches


class State(object):

    def __init__(self, args=None, opts=None):
        super(State, self).__init__()

        self.args = args if args is not None else []
        self.opts = opts

        self.typemap = {
            "tomio": {
                dict:       self.toDict,
                list:       self.toList,
                tuple:      self.toTuple,
                bool:       self.toBoolean,
                str:        "Bytes",
                bytes:      "Bytes",
                unicode:    "String",
                int:        "Number",
                type(None): "None",
                float:      "Number",
                Decimal:    "Number",
            },
            "frommio": {
                "Dict":    self.fromDict,
                "List":    self.fromList,
                "Tuple":   self.fromTuple,
                "Boolean": self.fromBoolean,
                "Bytes":   bytes,
                "String":  unicode,
                "Number":  float
            }
        }

        self.prompt = "mio> "

    def fromDict(self, x):
        return dict((self.frommio(k), self.frommio(v)) for k, v in x.value.items())

    def fromList(self, xs):
        return list(self.frommio(x) for x in xs.value)

    def fromTuple(self, xs):
        return tuple(self.frommio(x) for x in xs.value)

    def fromBoolean(self, x):
        if x.value is None:
            return None
        return bool(x.value)

    def toDict(self, x):
        return self.find("Dict").clone(dict(((self.tomio(k), self.tomio(v)) for k, v in x.items())))

    def toList(self, xs):
        return self.find("List").clone(list((self.tomio(x) for x in xs)))

    def toTuple(self, xs):
        return self.find("Tuple").clone(tuple((self.tomio(x) for x in xs)))

    def toBoolean(self, x):
        return self.find("True") if x else self.find("False")

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    def bootstrap(self):
        from .object import Object
        from .trait import Trait

        # Fake Object to solve chicken/egg problem.
        self.root = {"Object": None}

        object = Object()

        root = Object(methods=False)
        root.parent = object

        root["Object"] = object
        root["Trait"] = Trait()
        root["Root"] = root

        self.root = root

    def initialize(self):
        from .core import Core
        from .types import Types
        from .traits import Traits

        root = self.root

        root["Types"] = Types()
        root["Core"] = Core()
        root["Traits"] = Traits()

        # Bootstrap the system
        if self.opts is None or (self.opts is not None and not self.opts.nosys):
            self.eval("""Importer import("bootstrap")""")

        # Reset the last value
        del self.root["_"]

    def frommio(self, x, default=None):
        return self.typemap["frommio"].get(x.type, lambda x: default)(x)

    def tomio(self, x, default="None"):
        mapto = self.typemap["tomio"].get(type(x), default)

        if callable(mapto):
            return mapto(x)
        else:
            if mapto in ("True", "False", "None"):
                return self.find(mapto)
            return self.find(mapto).clone(x)

    def find(self, name):
        if "Types" in self.root and name in self.root["Types"]:
            return self.root["Types"][name]
        elif "Core" in self.root and name in self.root["Core"]:
            return self.root["Core"][name]
        elif "Traits" in self.root and name in self.root["Traits"]:
            return self.root["Traits"][name]
        elif "builtins" in self.root and name in self.root["builtins"]:
            return self.root["builtins"][name]
        else:
            return self.root[name]

    def eval(self, code, receiver=None, context=None, showerror=True, reraise=False):
        from .parser import parse
        from .lexer import tokenize

        message = None

        try:
            message = parse(tokenize(code))
            return message.eval(self.root if receiver is None else receiver, self.root if context is None else context, message)
        except Error as e:
            if showerror:
                print_traceback(e)
            if reraise:
                raise

    def load(self, filename, receiver=None, context=None):
        self.eval(open(filename, "r").read(), receiver=receiver, context=context)

    def runsource(self, source):
        from .utils import format_result

        if not check_parens(source):
            return True

        result = self.eval(source)
        if result is not None:
            output = format_result(result)
            if output is not None:
                print("===> {0:s}".format(output))

    def repl(self):
        from .utils import tryimport

        readline = tryimport("readline")

        if readline is not None:
            readline.parse_and_bind("tab: complete")

            completer = Completer(self)
            readline.set_completer_delims("")
            readline.set_completer(completer.complete)
            readline.set_completion_display_matches_hook(completer.display_matches)

        print("mio {0:s}".format(version))

        code = ""
        cont = False

        while True:
            try:
                if cont:
                    self.prompt = ".... "
                else:
                    self.prompt = "mio> "

                code += raw_input(self.prompt)

                if code:
                    if check_parens(code):
                        self.runsource(code)
                        code = ""
                        cont = False
                    else:
                        cont = True
            except EOFError:
                raise SystemExit(0)



def test(mio):
    mio.eval("i = 0")
    assert mio.eval("x = Continuation current; i += 1")
    assert mio.eval("i == 1")

    mio.eval("x()")
    assert mio.eval("i == 2")


def test2(mio):
    x = mio.eval("x = Continuation current")
    assert mio.eval("x") == x



import sys


from mio.object import Object
from mio.utils import default_repr, format_function, format_method, format_object, format_result, format_value, method, tryimport, Null


def foo():
    pass


def bar():
    "Return \"Bar\""


class Foo(Object):

    @method()
    def noargs(self, receiver, context, m):
        pass

    @method()
    def args(self, receiver, context, m, a, b, c):
        pass

    @method()
    def varargs(self, receiver, context, m, *args):
        pass


FOO_TEMPLATE = """Foo at {0:s}:
       args            = args(a, b, c)
       noargs          = noargs()
       varargs         = varargs(*args)"""


def test_format_object(mio):
    foo = Foo()
    assert format_object(foo) == FOO_TEMPLATE.format(hex(id(foo)))


def test_format_method(mio):
    foo = Foo()
    assert format_method(foo.noargs) == "noargs()"
    assert format_method(foo.args) == "args(a, b, c)"
    assert format_method(foo.varargs) == "varargs(*args)"


def test_format_function(mio):
    assert format_function(foo) == "foo()"
    assert format_function(bar) == "Return \"Bar\""


def test_format_value():
    obj = Foo()
    assert format_value(obj.noargs) == "noargs()"
    assert format_value(obj.args) == "args(a, b, c)"
    assert format_value(obj.varargs) == "varargs(*args)"


def test_format_value2():
    assert format_value(foo) == "foo()"
    assert format_value(bar) == "Return \"Bar\""


def test_format_value3():
    assert format_value(1) == "1"


def test_format_result():
    obj = Foo()
    assert format_result(obj.noargs) == "noargs()"
    assert format_result(obj.args) == "args(a, b, c)"
    assert format_result(obj.varargs) == "varargs(*args)"


def test_format_result2():
    assert format_result(foo) == "foo()"
    assert format_result(bar) == "Return \"Bar\""


def test_format_result3():
    assert format_result(1) == "1"


def test_format_result4(mio):
    assert format_result(mio.eval("None")) is None


def test_format_result5(mio):
    mio.eval("""
        Foo = Object clone() do(
            __repr__ = method("Foo")
        )
    """)

    assert format_result(mio.eval("Foo")) == "Foo"


def test_format_result6(mio):
    Foo = mio.eval("Foo = Object clone()")
    assert format_result(mio.eval("Foo")) == "Foo(Object) at {0:s}".format(hex(id(Foo)))


def test_default_repr(mio):
    Foo = mio.eval("Foo = Object clone()")
    assert default_repr(mio.eval("Foo")) == "Foo(Object) at {0:s}".format(hex(id(Foo)))


def test_tryimport(mio):
    m = tryimport("sys")
    assert m is sys


def test_tryimport_fail(mio):
    try:
        tryimport("foo", "foo")
    except Warning as w:
        assert w[0] == "foo"


def test_null(mio):
    assert Null is Null
    assert Null() is Null



from mio import runtime
from mio.utils import method
from mio.object import Object
from mio.errors import TypeError


class Trait(Object):

    def __init__(self):
        super(Trait, self).__init__()

        self.requirements = []

        self.create_methods()
        self.parent = runtime.find("Trait" if self.__class__ is not Trait else "Object")

    def __setitem__(self, key, value):
        if getattr(value, "type", None) != "Block":
            raise TypeError("Traits cannot contain state!")
        super(Trait, self).__setitem__(key, value)

    @method()
    def init(self, receiver, context, m):
        receiver.requirements = []
        return receiver

    @method()
    def requires(self, receiver, context, m, *methods):
        receiver.requirements.extend(list((runtime.state.frommio(method.eval(context)) for method in methods)))
        return receiver

    @method()
    def requirements(self, receiver, context, m):
        return runtime.state.tomio(receiver.requirements)



from pytest import raises

from mio.errors import IndexError


def test_range_init(mio):
    mio.eval("xs = List clone()") == []
    mio.eval("xs append(2)") == [2]
    mio.eval("xs append(8)") == [2, 8]
    mio.eval("xs append(2)") == [2, 8, 2]
    assert mio.eval("xs") == [2, 8, 2]

    assert mio.eval("list(Range clone(*xs))") == [2, 4, 6]


def test_range_start(mio):
    assert mio.eval("list(Range clone(3))") == [0, 1, 2]


def test_range_stop(mio):
    assert mio.eval("list(Range clone(2, 4))") == [2, 3]


def test_range_step(mio):
    assert mio.eval("list(Range clone(2, 8, 2))") == [2, 4, 6]


def test_range_setStep(mio):
    mio.eval("r = Range clone(2, 8, 2)")
    assert mio.eval("list(r)") == [2, 4, 6]

    mio.eval("r setStep(1)")
    assert mio.eval("list(r)") == [2, 3, 4, 5, 6, 7]


def test_range_invalid(mio):
    assert mio.eval("list(Range clone(10, 0, 1))") == []


def test_range_invalid2(mio):
    assert mio.eval("list(Range clone(0, 10, -1))") == []


def test_range_getitem(mio):
    mio.eval("r = Range clone(3)")
    assert mio.eval("r __getitem__(0)") == 0
    assert mio.eval("r __getitem__(1)") == 1
    assert mio.eval("r __getitem__(2)") == 2
    assert mio.eval("r __getitem__(3)") == 3


def test_range_getitem2(mio):
    mio.eval("r = Range clone(3)")
    assert mio.eval("r __getitem__(0)") == 0
    assert mio.eval("r __getitem__(1)") == 1
    assert mio.eval("r __getitem__(2)") == 2
    assert mio.eval("r __getitem__(3)") == 3

    with raises(IndexError):
        mio.eval("r __getitem__(4)", reraise=True)


def test_range_getitem3(mio):
    assert mio.eval("Range clone(10, 0, 1) __getitem__(0)") == 10


def test_range_repr(mio):
    assert repr(mio.eval("Range clone(3)")) == "range(3)"


def test_range_str(mio):
    assert str(mio.eval("Range clone(3)")) == "range(3)"



from pytest import raises


from mio.errors import StopIteration


def test_generator(mio):
    mio.eval("""foo = block(
        yield 1
        yield 2
        yield 3
    )""")

    mio.eval("f = foo()")
    assert mio.eval("next(f)") == 1
    assert mio.eval("next(f)") == 2
    assert mio.eval("next(f)") == 3

    with raises(StopIteration):
        mio.eval("next(f)", reraise=True)



from mio import runtime
from mio.trait import Trait
from mio.utils import method
from mio.errors import TypeError


class TAdaptable(Trait):

    def __init__(self):
        super(TAdaptable, self).__init__()

        Object = runtime.find("Object")
        Object.__addtrait__(self)

    @method()
    def use(self, receiver, context, m, trait, resolution=None):
        trait = trait.eval(context)

        if receiver.__hastrait__(trait):
            raise TypeError("{0:s} already uses {1:s}".format(repr(receiver), repr(trait)))

        resolution = runtime.state.frommio(resolution.eval(context)) if resolution is not None else {}

        receiver.__addtrait__(trait, **resolution)

        return receiver

    @method()
    def delTrait(self, receiver, context, m, trait):
        trait = trait.eval(context)

        if not receiver.__hastrait__(trait):
            raise TypeError("{0:s} does not use {1:s}".format(repr(receiver), repr(trait)))

        receiver.__deltrait__(trait)

        return receiver

    @method()
    def hasTrait(self, receiver, context, m, trait):
        trait = trait.eval(context)
        return runtime.find("True") if receiver.__hastrait__(trait) else runtime.find("False")

    @method("traits", True)
    def getTraits(self, receiver, context, m):
        return runtime.find("List").clone(receiver.traits.keys())

    @method("behaviors", True)
    def getBehaviors(self, receiver, context, m):
        return runtime.find("List").clone(receiver.behaviors.keys())

    @method()
    def adapt(self, receiver, context, m, trait):
        trait = trait.eval(context)
        obj = receiver.clone()
        obj.__addtrait__(trait)
        return obj



from pytest import raises


from mio.errors import AttributeError, TypeError


def test_basic_trait(mio, capfd):
    mio.eval("""
        TGreeting = Trait clone() do (
            hello = method(
                print("Hello", self getGreeting())
            )
        )

        World = Object clone() do (
            use(TGreeting)

            greeting = "World!"

            getGreeting = method(
               self greeting
            )

            setGreeting = method(aGreeting,
                self greeting = aGreeting
            )
        )
    """)

    mio.eval("World hello()")
    out, err = capfd.readouterr()
    assert out == "Hello World!\n"

    mio.eval("World setGreeting(\"John\")")
    mio.eval("World hello()")
    out, err = capfd.readouterr()
    assert out == "Hello John\n"


def test_basic_trait2(mio, capfd):
    mio.eval("""
        TGreeting = Trait clone() do (
            hello = method(
                print("Hello", self getGreeting())
            )
        )

        World = Object clone() do (
            use(TGreeting)

            greeting = "World!"

            getGreeting = method(
               self greeting
            )

            setGreeting = method(aGreeting,
                self greeting = aGreeting
            )
        )
    """)

    mio.eval("World hello()")
    out, err = capfd.readouterr()
    assert out == "Hello World!\n"

    mio.eval("World setGreeting(\"John\")")
    mio.eval("World hello()")
    out, err = capfd.readouterr()
    assert out == "Hello John\n"

    with raises(TypeError):
        mio.eval("World use(TGreeting)", reraise=True)


def test_invalid(mio):
    mio.eval("TGreetable = Object clone()")

    with raises(TypeError):
        mio.eval("Object clone() use(TGreetable)", reraise=True)


def test_state(mio):
    mio.eval("TGreetable = Trait clone()")

    with raises(TypeError):
        mio.eval("TGreetable greeting = \"World!\"", reraise=True)


def test_requirements(mio):
    mio.eval("""
        TGreetable = Trait clone() do(
            requires("greeting")
        )
    """)

    mio.eval("TGreetable requirements()") == [u"greeting"]


def test_requires(mio):
    mio.eval("""
        TGreetable = Trait clone() do(
            requires("greeting")
        )
    """)

    with raises(TypeError):
        mio.eval("Object clone() use(TGreetable)", reraise=True)


def test_resolution(mio):
    mio.eval("""
        TFoo = Trait clone() do(
            foo = method("foo")
        )
        TBar = Trait clone() do(
            foo = method("foo")
        )
    """)

    with raises(TypeError):
        mio.eval("Object clone() use(TFoo) use(TBar)", reraise=True)


def test_resolution2(mio):
    mio.eval("""
        TFoo = Trait clone() do(
            foo = method("foo")
        )
        TBar = Trait clone() do(
            foo = method("foo")
        )
    """)

    mio.eval("Foo = Object clone() use(TFoo) use(TBar, {\"foo\": \"bar\"})")
    assert mio.eval("Foo hasTrait(TFoo)")
    assert mio.eval("Foo hasTrait(TBar)")
    assert mio.eval("Foo behaviors") == ["foo", "bar"]


def test_resolution_deltrait(mio):
    mio.eval("""
        TFoo = Trait clone() do(
            foo = method("foo")
        )
        TBar = Trait clone() do(
            foo = method("foo")
        )
    """)

    mio.eval("Foo = Object clone() use(TFoo) use(TBar, {\"foo\": \"bar\"})")
    assert mio.eval("Foo hasTrait(TFoo)")
    assert mio.eval("Foo hasTrait(TBar)")
    assert mio.eval("Foo behaviors") == ["foo", "bar"]

    mio.eval("Foo delTrait(TFoo)")
    assert not mio.eval("Foo hasTrait(TFoo)")
    assert mio.eval("Foo behaviors") == ["bar"]

    mio.eval("Foo delTrait(TBar)")
    assert not mio.eval("Foo hasTrait(TBar)")
    assert mio.eval("Foo behaviors") == []


def test_adapt(mio):
    mio.eval("TGreetable = Trait clone()")
    assert mio.eval("World = Object clone() adapt(TGreetable) hasTrait(TGreetable)")


def test_hasTrait(mio):
    mio.eval("""
        TGreetable = Trait clone()
        World = Object clone() do (
            use(TGreetable)
        )
    """)

    assert mio.eval("World hasTrait(TGreetable)")


def test_hasTrait2(mio):
    mio.eval("""
        TGreetable = Trait clone()
        World = Object clone() do (
            use(TGreetable)
        )
    """)

    assert mio.eval("World hasTrait(TGreetable)")
    assert mio.eval("World clone() hasTrait(TGreetable)")


def test_delTrait(mio):
    mio.eval("""
        TGreetable = Trait clone()
        World = Object clone() do (
            use(TGreetable)
        )
    """)

    assert mio.eval("World hasTrait(TGreetable)")
    mio.eval("World delTrait(TGreetable)")
    assert mio.eval("World behaviors") == []
    assert not mio.eval("World hasTrait(TGreetable)")


def test_delTrait2(mio, capfd):
    mio.eval("""
        TGreetable = Trait clone() do (
            hello = method(
                print("Hello World!")
            )
        )

        World = Object clone() do (
            use(TGreetable)
        )
    """)

    assert mio.eval("World hasTrait(TGreetable)")
    assert mio.eval("World behaviors") == ["hello"]

    assert mio.eval("World hello()").value is None
    out, err = capfd.readouterr()
    assert out == "Hello World!\n"

    mio.eval("World delTrait(TGreetable)")
    assert mio.eval("World behaviors") == []
    assert not mio.eval("World hasTrait(TGreetable)")

    with raises(AttributeError):
        mio.eval("World hello()", reraise=True)


def test_delTrait3(mio, capfd):
    mio.eval("""
        TGreetable = Trait clone() do (
            hello = method(
                print("Hello World!")
            )
        )

        World = Object clone() do (
            use(TGreetable)
        )
    """)

    assert mio.eval("World hasTrait(TGreetable)")
    assert mio.eval("World behaviors") == ["hello"]

    assert mio.eval("World hello()").value is None
    out, err = capfd.readouterr()
    assert out == "Hello World!\n"

    mio.eval("World delTrait(TGreetable)")
    assert mio.eval("World behaviors") == []
    assert not mio.eval("World hasTrait(TGreetable)")

    with raises(TypeError):
        mio.eval("World delTrait(TGreetable)", reraise=True)


def test_traits(mio):
    mio.eval("""
        TGreetable = Trait clone()
        World = Object clone() do (
            use(TGreetable)
        )
    """)

    TGreetable = mio.eval("TGreetable")
    assert mio.eval("World traits") == [TGreetable]


def test_behaviors(mio, capfd):
    mio.eval("""
        TGreetable = Trait clone() do (
            hello = method(
                print("Hello World!")
            )
        )

        World = Object clone() do (
            use(TGreetable)
        )
    """)

    assert mio.eval("World hello()").value is None
    out, err = capfd.readouterr()
    assert out == "Hello World!\n"
    assert mio.eval("World behaviors") == ["hello"]


def test_del_behavior(mio, capfd):
    mio.eval("""
        TGreetable = Trait clone() do (
            hello = method(
                print("Hello World!")
            )
        )

        World = Object clone() do (
            use(TGreetable)
        )
    """)

    assert mio.eval("World hello()").value is None
    out, err = capfd.readouterr()
    assert out == "Hello World!\n"
    assert mio.eval("World behaviors") == ["hello"]

    mio.eval("World del(\"hello\")")

    with raises(AttributeError):
        mio.eval("World hello()", reraise=True)
    assert mio.eval("World behaviors") == []



from copy import copy
from operator import xor
from collections import OrderedDict
from inspect import getmembers, ismethod


from mio import runtime
from mio.utils import method, Null
from mio.errors import AttributeError, TypeError


class Object(object):

    __slots__ = ("attrs", "binding", "parent", "state", "value", "traits", "behaviors",)

    def __init__(self, value=Null, methods=True):
        super(Object, self).__init__()

        self.attrs = {}
        self.parent = None
        self.value = value

        self.traits = {}
        self.behaviors = {}

        self.binding = None

        from mio.core.state import State
        if self.__class__ is not State:
            self.state = State()

        if methods:
            self.create_methods()

    def create_methods(self):
        keys = self.__class__.__dict__.keys()
        self.attrs.update(((v.name, v) for k, v in getmembers(self, ismethod) if getattr(v, "method", False) and k in keys))

    def __hash__(self):
        return (
            reduce(xor, map(hash, self.attrs.keys()), 0) ^
            hash(self.parent) ^
            hash(self.value) ^
            reduce(xor, map(hash, self.traits.keys()), 0) ^
            reduce(xor, map(hash, self.behaviors.keys()), 0) ^
            hash(self.binding)
        )

    def __nonzero__(self):
        return bool(self.value)

    def __eq__(self, other):
        return self.value == getattr(other, "value", other)

    def __cmp__(self, other):
        return cmp(self.value, getattr(other, "value", other))

    def __contains__(self, key):
        return key in self.attrs or key in self.behaviors

    def __delitem__(self, key):
        if key in self.attrs:
            del self.attrs[key]
        elif key in self.behaviors:
            del self.behaviors[key]

    def __getitem__(self, key):
        if key in self.attrs:
            return self.attrs[key]
        elif key in self.behaviors:
            return self.behaviors[key]
        else:
            parent = self.parent
            while parent is not None:
                if key in parent:
                    return parent[key]
                parent = parent.parent

        try:
            return self.forward(key)
        except:
            raise AttributeError("{0:s} has no attribute {1:s}".format(self.type, repr(key)))

    def __setitem__(self, key, value):
        self.attrs[key] = value

    def __addtrait__(self, trait, **resolution):
        from .trait import Trait

        if not isinstance(trait, Trait):
            raise TypeError("Trait expected but got {0:s}".format(repr(getattr(trait, "type", trait.__class__.__name__))))

        for requirement in trait.requirements:
            if not self.lookup(requirement):
                raise TypeError("{0:s} requires {0:s}".format(repr(trait), repr(requirement)))

        for k, v in trait.attrs.items():
            if k in self:
                name = resolution.get(k, None)
                if not name:
                    raise TypeError("Method {0:s} of {1:s} conflicts with {2:s}".format(k, repr(trait), repr(self)))
                else:
                    self.behaviors[name] = v
            else:
                self.behaviors[k] = v

        self.traits[trait] = resolution

    def __deltrait__(self, trait):
        resolution = self.traits[trait]

        for k, v in trait.attrs.items():
            if k in self.behaviors:
                del self.behaviors[k]
            else:
                del self.behaviors[resolution[k]]

        del self.traits[trait]

    def __hastrait__(self, trait):
        if trait in self.traits:
            return True

        parent = self.parent
        while parent is not None:
            if trait in parent.traits:
                return True
            parent = parent.parent

        return False

    def __repr__(self):
        type = "{0:s}({1:s})".format(self.binding, self.type) if self.binding is not None else self.type
        default = "{0:s} at {1:s}".format(type, hex(id(self)))
        return repr(self.value) if self.value is not Null else default

    def lookup(self, name):
        try:
            self[name]
            return True
        except AttributeError:
            return False

    def clone(self, value=Null):
        obj = copy(self)

        obj.attrs = {}
        obj.parent = self
        obj.value = value if value is not Null else obj.value

        obj.traits = {}
        obj.behaviors = {}

        self.binding = None

        from mio.core.state import State
        if self.__class__ is not State:
            self.state = State()

        return obj

    def forward(self, key):
        return runtime.find(key)

    @property
    def type(self):
        return self.__class__.__name__

    # Attribute Operations

    @method("del")
    def _del(self, receiver, context, m, key):
        key = unicode(key.eval(context))
        value = receiver[key]
        del receiver[key]
        if isinstance(value, Object):
            value.bindig = None
        return runtime.find("None")

    @method()
    def has(self, receiver, context, m, key):
        key = unicode(key.eval(context))
        if key in receiver:
            return runtime.find("True")
        return runtime.find("False")

    @method()
    def set(self, receiver, context, m, key, value):
        key = unicode(key.eval(context))
        value = value.eval(context)
        if isinstance(value, Object):
            value.binding = key
        receiver[key] = value
        return value

    @method()
    def get(self, receiver, context, m, key, default=None):
        key = unicode(key.eval(context))
        default = default.eval(context) if default else runtime.find("None")
        return receiver.attrs.get(key, default)

    # Block Operations

    @method()
    def block(self, receiver, context, m, *args):
        args, body = args[:-1], args[-1:][0]

        # Evaluate kwargs first
        ctx = runtime.find("Object").clone()
        kwargs = OrderedDict([(arg.args[0].name, arg.eval(ctx)) for arg in args if arg.name == "set"])

        args = [arg for arg in args if not arg.name == "set"]

        from mio.core.block import Block
        return Block(body, args, kwargs, context)

    @method("method")
    def _method(self, receiver, context, m, *args):
        args, body = args[:-1], args[-1:][0]

        # Evaluate kwargs first
        ctx = runtime.find("Object").clone()
        kwargs = OrderedDict([(arg.args[0].name, arg.eval(ctx)) for arg in args if arg.name == "set"])

        args = [arg for arg in args if not arg.name == "set"]

        from mio.core.block import Block
        return Block(body, args, kwargs)

    # Introspection

    @method("state", True)
    def getState(self, receiver, context, m):
        return receiver.state

    @method("type", True)
    def getType(self, receiver, context, m):
        from mio.types import String
        return String(receiver.type)

    @method("parent", True)
    def _parent(self, receiver, context, m):
        if receiver.parent is not None:
            return receiver.parent
        return receiver

    @method("value", True)
    def _value(self, receiver, context, m):
        if receiver.value is not Null:
            return receiver.value
        return runtime.find("None")

    @method("__hash__", property=True)
    def hash(self, receiver, context, m):
        try:
            return runtime.find("Number").clone(hash(receiver))
        except:
            return runtime.find("None")

    @method(property=True)
    def id(self, receiver, context, m, obj=None):
        obj = obj.eval(context) if obj is not None else receiver
        return runtime.find("Number").clone(id(obj))

    @method(property=True)
    def keys(self, receiver, context, m):
        String = runtime.find("String")
        keys = [String.clone(key) for key in receiver.attrs.keys()]
        return runtime.find("List").clone(keys)

    # Object Operations

    @method()
    def init(self, receiver, context, m):
        return receiver

    @method()
    def primitive(self, receiver, context, m, method, *args):
        method = unicode(method.eval(context))
        args = [arg.eval(context).value for arg in args]
        if hasattr(receiver, method):
            return getattr(receiver, method)(*args)
        raise AttributeError("{0:s} has no attribute {1:s}".format(receiver.type, repr(method)))

    @method()
    def evalArg(self, receiver, context, m, *args):
        if len(args) > 1:
            return runtime.find("Tuple").clone(tuple(arg.eval(context) for arg in args))
        elif len(args) == 1:
            return args[0].eval(context)
        else:
            return runtime.find("None")

    @method()
    def do(self, receiver, context, m, expression):
        expression.eval(receiver)
        return receiver

    @method()
    def setParent(self, receiver, context, m, parent):
        parent = parent.eval(context)
        if parent is receiver:
            raise TypeError("Canoot set parent to self!")

        receiver.parent = parent

        return receiver

    @method()
    def setValue(self, receiver, context, m, value):
        receiver.value = value.eval(context)
        return receiver

    # Boolean Operations

    @method()
    def cmp(self, receiver, context, m, other):
        return runtime.find("Number").clone(cmp(receiver, other.eval(context)))

    # Type Conversion

    @method("__repr__")
    def getRepr(self, receiver, context, m):
        return runtime.find("String").clone(repr(receiver))

    @method("__str__")
    def getStr(self, receiver, context, m):
        return runtime.find("String").clone(unicode(receiver))
